@@ -14,25 +14,30 @@ Thank you, Kurt Jacobson!
 """
 
 import argparse
-import os.path
 import sys
-
+import threading
 import gi
+from nwg_displays.settings_applier import SettingsApplier
+from nwg_displays.wallpaper_manager import WallpaperManager
 
-gi.require_version('Gtk', '3.0')
+
+gi.require_version("Gtk", "3.0")
 try:
-    gi.require_version('GtkLayerShell', '0.1')
+    gi.require_version("GtkLayerShell", "0.1")
 except ValueError:
-    raise RuntimeError('\n\n' +
-                       'If you haven\'t installed GTK Layer Shell, you need to point Python to the\n' +
-                       'library by setting GI_TYPELIB_PATH and LD_LIBRARY_PATH to <build-dir>/src/.\n' +
-                       'For example you might need to run:\n\n' +
-                       'GI_TYPELIB_PATH=build/src LD_LIBRARY_PATH=build/src python3 ' + ' '.join(sys.argv))
+    raise RuntimeError(
+        "\n\n"
+        + "If you haven't installed GTK Layer Shell, you need to point Python to the\n"
+        + "library by setting GI_TYPELIB_PATH and LD_LIBRARY_PATH to <build-dir>/src/.\n"
+        + "For example you might need to run:\n\n"
+        + "GI_TYPELIB_PATH=build/src LD_LIBRARY_PATH=build/src python3 "
+        + " ".join(sys.argv)
+    )
 
 from gi.repository import Gtk, GLib, GtkLayerShell
 
 from nwg_displays.tools import *
-
+from nwg_displays.profiles import ProfileManager
 from nwg_displays.__about__ import __version__
 
 dir_name = os.path.dirname(__file__)
@@ -45,12 +50,20 @@ old_config_dir = os.path.join(get_config_home(), "nwg-outputs")
 
 sway_config_dir = os.path.join(get_config_home(), "sway")
 if sway and not os.path.isdir(sway_config_dir):
-    print("WARNING: Couldn't find sway config directory '{}'".format(sway_config_dir), file=sys.stderr)
+    print(
+        "[Warning] Couldn't find sway config directory '{}'".format(sway_config_dir),
+        file=sys.stderr,
+    )
     sys.exit(1)
 
 hypr_config_dir = os.path.join(get_config_home(), "hypr")
 if hypr and not os.path.isdir(hypr_config_dir):
-    print("WARNING: Couldn't find Hyprland config directory '{}'".format(hypr_config_dir), file=sys.stderr)
+    print(
+        "[Warning] Couldn't find Hyprland config directory '{}'".format(
+            hypr_config_dir
+        ),
+        file=sys.stderr,
+    )
     sys.exit(1)
 
 # Create empty files if not found
@@ -61,7 +74,7 @@ elif hypr:
     for name in ["monitors.conf", "workspaces.conf"]:
         create_empty_file(os.path.join(hypr_config_dir, name))
 else:
-    eprint("Neither sway nor Hyprland detected, terminating")
+    eprint("[Error] Neither sway nor Hyprland detected, terminating")
     sys.exit(1)
 
 config = {}
@@ -73,12 +86,15 @@ i3.get_outputs() does not return some output attributes, especially when connect
 i3.get_tree() on the other hand does not return inactive outputs. So we'll list attributes with .get_tree(),
 and the add inactive outputs, if any, from what we detect with .get_outputs()
 """
-outputs = {}  # Active outputs, listed from the sway tree; stores name and all attributes.
+outputs = (
+    {}
+)  # Active outputs, listed from the sway tree; stores name and all attributes.
 outputs_activity = {}  # Just a dictionary "name": is_active - from get_outputs()
 workspaces = {}  # "workspace_num": "display_name"
 
 display_buttons = []
 selected_output_button = None
+profile_manager = None
 
 # Glade form fields
 form_name = None
@@ -104,6 +120,7 @@ form_apply = None
 form_version = None
 form_mirror = None
 form_ten_bit = None
+form_profile_wallpapers = None
 
 dialog_win = None
 confirm_win = None
@@ -150,7 +167,11 @@ def load_vocabulary():
     if lang is None:
         lang = "en_US"
     else:
-        lang = lang.split(".")[0] if not shell_data["interface-locale"] else shell_data["interface-locale"]
+        lang = (
+            lang.split(".")[0]
+            if not shell_data["interface-locale"]
+            else shell_data["interface-locale"]
+        )
 
     # translate if translation available
     if lang != "en_US":
@@ -187,8 +208,12 @@ def on_button_press_event(widget, event):
         # note that we're rounding down now so that these max values don't get
         # rounded upward later and push the widget off the edge of its parent.
         global max_x, max_y
-        max_x = round_down_to_multiple(p.get_allocation().width - widget.get_allocation().width, SENSITIVITY)
-        max_y = round_down_to_multiple(p.get_allocation().height - widget.get_allocation().height, SENSITIVITY)
+        max_x = round_down_to_multiple(
+            p.get_allocation().width - widget.get_allocation().width, SENSITIVITY
+        )
+        max_y = round_down_to_multiple(
+            p.get_allocation().height - widget.get_allocation().height, SENSITIVITY
+        )
 
         update_form_from_widget(widget)
 
@@ -214,6 +239,7 @@ def on_motion_notify_event(widget, event):
         px = x
         py = y
         snap_x, snap_y = [0], [0]
+        # Collect snap lines from other displays
         for db in display_buttons:
             if db.name == widget.name:
                 continue
@@ -235,6 +261,7 @@ def on_motion_notify_event(widget, event):
                 snap_y.append(val)
 
         snap_h, snap_v = None, None
+        # Find nearest horizontal snap line
         for value in snap_x:
             if abs(x - value) < snap_threshold_scaled:
                 snap_h = value
@@ -246,6 +273,7 @@ def on_motion_notify_event(widget, event):
                 snap_h = value - w
                 break
 
+        # Find nearest vertical snap line
         for value in snap_y:
             if abs(y - value) < snap_threshold_scaled:
                 snap_v = value
@@ -297,7 +325,11 @@ def update_form_from_widget(widget):
     form_dpms.set_active(widget.dpms)
     form_adaptive_sync.set_active(widget.adaptive_sync)
     form_custom_mode.set_active(widget.custom_mode)
-    form_view_scale.set_value(config["view-scale"])  # not really from the widget, but from the global value
+    form_view_scale.set_value(
+        config["view-scale"]
+    )  # not really from the widget, but from the global value
+    if form_profile_wallpapers:
+        form_profile_wallpapers.set_active(config.get("profile-bound-wallpapers", True))
     form_use_desc.set_active(config["use-desc"])
     form_x.set_value(widget.x)
     form_y.set_value(widget.y)
@@ -323,13 +355,21 @@ def update_form_from_widget(widget):
     form_modes.remove_all()
     active = ""
     for mode in widget.modes:
-        m = "{}x{}@{}Hz".format(mode["width"], mode["height"], mode["refresh"] / 1000, mode[
-            "refresh"] / 1000, widget.refresh)
+        m = "{}x{}@{}Hz".format(
+            mode["width"],
+            mode["height"],
+            mode["refresh"] / 1000,
+            mode["refresh"] / 1000,
+            widget.refresh,
+        )
         form_modes.append(m, m)
         # This is just to set active_id
 
-        if mode["width"] == widget.physical_width and mode["height"] == widget.physical_height and mode[
-            "refresh"] / 1000 == widget.refresh:
+        if (
+            mode["width"] == widget.physical_width
+            and mode["height"] == widget.physical_height
+            and mode["refresh"] / 1000 == widget.refresh
+        ):
             active = m
     if active:
         form_modes.set_active_id(active)
@@ -340,9 +380,28 @@ def update_form_from_widget(widget):
 
 
 class DisplayButton(Gtk.Button):
-    def __init__(self, name, description, x, y, physical_width, physical_height, transform, scale, scale_filter,
-                 refresh, modes, active, dpms, adaptive_sync_status, ten_bit, custom_mode_status, focused, monitor,
-                 mirror=""):
+    def __init__(
+        self,
+        name,
+        description,
+        x,
+        y,
+        physical_width,
+        physical_height,
+        transform,
+        scale,
+        scale_filter,
+        refresh,
+        modes,
+        active,
+        dpms,
+        adaptive_sync_status,
+        ten_bit,
+        custom_mode_status,
+        focused,
+        monitor,
+        mirror="",
+    ):
         super().__init__()
         # Output properties
         self.name = name
@@ -362,7 +421,9 @@ class DisplayButton(Gtk.Button):
         # self.modes = modes
         self.active = active
         self.dpms = dpms
-        self.adaptive_sync = adaptive_sync_status == "enabled"  # converts "enabled | disabled" to bool
+        self.adaptive_sync = (
+            adaptive_sync_status == "enabled"
+        )  # converts "enabled | disabled" to bool
         self.custom_mode = custom_mode_status
         self.focused = focused
         self.mirror = mirror
@@ -380,8 +441,13 @@ class DisplayButton(Gtk.Button):
         self.rescale_transform()
         self.set_property("name", "output")
 
-        self.indicator = Indicator(monitor, name, round(self.physical_width * config["view-scale"]),
-                                   round(self.physical_height * config["view-scale"]), config["indicator-timeout"])
+        self.indicator = Indicator(
+            monitor,
+            name,
+            round(self.physical_width * config["view-scale"]),
+            round(self.physical_height * config["view-scale"]),
+            config["indicator-timeout"],
+        )
 
         self.show()
 
@@ -406,11 +472,14 @@ class DisplayButton(Gtk.Button):
         selected_output_button = self
 
     def unselect(self):
+        self.selected = False
         self.set_property("name", "output")
 
     def rescale_transform(self):
-        self.set_size_request(round(self.logical_width * config["view-scale"]),
-                              round(self.logical_height * config["view-scale"]))
+        self.set_size_request(
+            round(self.logical_width * config["view-scale"]),
+            round(self.logical_height * config["view-scale"]),
+        )
 
     def on_active_check_button_toggled(self, w):
         self.active = w.get_active()
@@ -436,6 +505,11 @@ def on_view_scale_changed(*args):
     save_json(config, os.path.join(config_dir, "config"))
 
 
+def on_profile_wallpapers_toggled(widget):
+    config["profile-bound-wallpapers"] = widget.get_active()
+    save_json(config, os.path.join(config_dir, "config"))
+
+
 def on_transform_changed(*args):
     if selected_output_button:
         transform = form_transform.get_active_id()
@@ -478,15 +552,21 @@ def on_custom_mode_toggle(widget):
 def on_pos_x_changed(widget):
     if selected_output_button:
         selected_output_button.x = round(widget.get_value())
-        fixed.move(selected_output_button, selected_output_button.x * config["view-scale"],
-                   selected_output_button.y * config["view-scale"])
+        fixed.move(
+            selected_output_button,
+            selected_output_button.x * config["view-scale"],
+            selected_output_button.y * config["view-scale"],
+        )
 
 
 def on_pos_y_changed(widget):
     if selected_output_button:
         selected_output_button.y = round(widget.get_value())
-        fixed.move(selected_output_button, selected_output_button.x * config["view-scale"],
-                   selected_output_button.y * config["view-scale"])
+        fixed.move(
+            selected_output_button,
+            selected_output_button.x * config["view-scale"],
+            selected_output_button.y * config["view-scale"],
+        )
 
 
 def on_width_changed(widget):
@@ -535,9 +615,27 @@ def on_mirror_selected(widget):
         selected_output_button.mirror = widget.get_active_id()
 
 
-def on_apply_button(widget):
+def on_apply_button(widget, p_manager=None):
     global outputs_activity
-    apply_settings(display_buttons, outputs_activity, outputs_path, use_desc=config["use-desc"])
+
+    if p_manager is None:
+        global profile_manager
+        p_manager = profile_manager
+
+    profile_name = p_manager.current_profile if p_manager else None
+    # Remove legacy storage
+    if "active-profile" in config:
+        del config["active-profile"]
+
+    SettingsApplier.apply_from_gui(
+        display_buttons,
+        outputs_activity,
+        outputs_path,
+        config["use-desc"],
+        create_confirm_win,
+        config_dir,
+        profile_name=profile_name,
+    )
     # save config file
     save_json(config, os.path.join(config_dir, "config"))
 
@@ -570,15 +668,35 @@ def create_display_buttons():
     for key in outputs:
         item = outputs[key]
         custom_mode = key in config["custom-mode"]
-        b = DisplayButton(key, item["description"], item["x"], item["y"], round(item["physical-width"]),
-                          round(item["physical-height"]),
-                          item["transform"], item["scale"], item["scale_filter"], item["refresh"], item["modes"],
-                          item["active"], item["dpms"], item["adaptive_sync_status"], item["ten_bit"], custom_mode,
-                          item["focused"], item["monitor"], mirror=item["mirror"])
+        b = DisplayButton(
+            key,
+            item["description"],
+            item["x"],
+            item["y"],
+            round(item["physical-width"]),
+            round(item["physical-height"]),
+            item["transform"],
+            item["scale"],
+            item["scale_filter"],
+            item["refresh"],
+            item["modes"],
+            item["active"],
+            item["dpms"],
+            item["adaptive_sync_status"],
+            item["ten_bit"],
+            custom_mode,
+            item["focused"],
+            item["monitor"],
+            mirror=item["mirror"],
+        )
 
         display_buttons.append(b)
 
-        fixed.put(b, round(item["x"] * config["view-scale"]), round(item["y"] * config["view-scale"]))
+        fixed.put(
+            b,
+            round(item["x"] * config["view-scale"]),
+            round(item["y"] * config["view-scale"]),
+        )
 
     display_buttons[0].select()
     update_form_from_widget(display_buttons[0])
@@ -624,7 +742,9 @@ def handle_keyboard(window, event):
 def create_workspaces_window(btn):
     global sway_config_dir
     global workspaces
-    workspaces = load_workspaces(os.path.join(sway_config_dir, "workspaces"), use_desc=config["use-desc"])
+    workspaces = load_workspaces(
+        os.path.join(sway_config_dir, "workspaces"), use_desc=config["use-desc"]
+    )
     old_workspaces = workspaces.copy()
     global dialog_win
     if dialog_win:
@@ -665,7 +785,9 @@ def create_workspaces_window(btn):
     btn_apply = Gtk.Button()
     btn_apply.set_label(voc["apply"])
     if sway_config_dir:
-        btn_apply.connect("clicked", on_workspaces_apply_btn, dialog_win, old_workspaces)
+        btn_apply.connect(
+            "clicked", on_workspaces_apply_btn, dialog_win, old_workspaces
+        )
     else:
         btn_apply.set_sensitive(False)
         btn_apply.set_tooltip_text("Config dir not found")
@@ -682,7 +804,8 @@ def create_workspaces_window(btn):
 def create_workspaces_window_hypr(btn):
     global workspaces
     workspaces = load_workspaces_hypr(
-        os.path.join(hypr_config_dir, "workspaces.conf"), num_ws=num_ws)
+        os.path.join(hypr_config_dir, "workspaces.conf"), num_ws=num_ws
+    )
     eprint("WS->Mon:", workspaces)
     old_workspaces = workspaces.copy()
     global dialog_win
@@ -703,7 +826,9 @@ def create_workspaces_window_hypr(btn):
     for i in range(num_ws):
         lbl = Gtk.Label()
         if config["use-desc"]:
-            lbl.set_markup("Workspace rule: <b>workspace={},monitor:desc:</b>".format(i + 1))
+            lbl.set_markup(
+                "Workspace rule: <b>workspace={},monitor:desc:</b>".format(i + 1)
+            )
         else:
             lbl.set_markup("Workspace rule: <b>workspace={},monitor:</b>".format(i + 1))
         lbl.set_property("halign", Gtk.Align.END)
@@ -729,7 +854,9 @@ def create_workspaces_window_hypr(btn):
     btn_apply = Gtk.Button()
     btn_apply.set_label(voc["apply"])
     if hypr_config_dir:
-        btn_apply.connect("clicked", on_workspaces_apply_btn_hypr, dialog_win, old_workspaces)
+        btn_apply.connect(
+            "clicked", on_workspaces_apply_btn_hypr, dialog_win, old_workspaces
+        )
     else:
         btn_apply.set_sensitive(False)
         btn_apply.set_tooltip_text("Config dir not found")
@@ -755,7 +882,11 @@ def close_dialog(w, win):
 def on_workspaces_apply_btn(w, win, old_workspaces):
     global workspaces
     if workspaces != old_workspaces:
-        save_workspaces(workspaces, os.path.join(sway_config_dir, "workspaces"), use_desc=config["use-desc"])
+        save_workspaces(
+            workspaces,
+            os.path.join(sway_config_dir, "workspaces"),
+            use_desc=config["use-desc"],
+        )
         notify("Workspaces assignment", "Restart sway for changes to take effect")
 
     close_dialog(w, win)
@@ -768,9 +899,12 @@ def on_workspaces_apply_btn_hypr(w, win, old_workspaces):
         text_file = open(workspace_conf_file, "w")
 
         now = datetime.datetime.now()
-        line = "# Generated by nwg-displays on {} at {}. Do not edit manually.\n".format(
-            datetime.datetime.strftime(now, '%Y-%m-%d'),
-            datetime.datetime.strftime(now, '%H:%M:%S'))
+        line = (
+            "# Generated by nwg-displays on {} at {}. Do not edit manually.\n".format(
+                datetime.datetime.strftime(now, "%Y-%m-%d"),
+                datetime.datetime.strftime(now, "%H:%M:%S"),
+            )
+        )
         text_file.write(line + "\n")
 
         monitors_with_default_workspace = []
@@ -793,120 +927,7 @@ def on_workspaces_apply_btn_hypr(w, win, old_workspaces):
     close_dialog(w, win)
 
 
-def apply_settings(display_buttons, outputs_activity, outputs_path, use_desc=False):
-    now = datetime.datetime.now()
-    lines = ["# Generated by nwg-displays on {} at {}. Do not edit manually.\n".format(
-        datetime.datetime.strftime(now, '%Y-%m-%d'),
-        datetime.datetime.strftime(now, '%H:%M:%S'))]
-    cmds = []
-    db_names = []
-    # just active outputs have their buttons
-    if os.getenv("SWAYSOCK"):
-        for db in display_buttons:
-            name = db.name if not use_desc else db.description
-            db_names.append(name)
-
-            lines.append('output "%s" {' % name)
-            cmd = 'output "{}"'.format(name)
-
-            custom_mode_str = "--custom" if db.custom_mode else ""
-            lines.append(
-                "    mode {} {}x{}@{}Hz".format(custom_mode_str, db.physical_width, db.physical_height, db.refresh))
-            cmd += " mode {} {}x{}@{}Hz".format(custom_mode_str, db.physical_width, db.physical_height, db.refresh)
-
-            lines.append("    pos {} {}".format(db.x, db.y))
-            cmd += " pos {} {}".format(db.x, db.y)
-
-            lines.append("    transform {}".format(db.transform))
-            cmd += " transform {}".format(db.transform)
-
-            lines.append("    scale {}".format(db.scale))
-            cmd += " scale {}".format(db.scale)
-
-            lines.append("    scale_filter {}".format(db.scale_filter))
-            cmd += " scale_filter {}".format(db.scale_filter)
-
-            a_s = "on" if db.adaptive_sync else "off"
-            lines.append("    adaptive_sync {}".format(a_s))
-            cmd += " adaptive_sync {}".format(a_s)
-
-            dpms = "on" if db.dpms else "off"
-            lines.append("    dpms {}".format(dpms))
-            cmd += " dpms {}".format(dpms)
-
-            lines.append("}")
-            cmds.append(cmd)
-
-        if not use_desc:
-            for key in outputs_activity:
-                if key not in db_names:
-                    lines.append('output "{}" disable'.format(key))
-                    cmds.append('output "{}" disable'.format(key))
-        else:
-            for key in outputs_activity:
-                desc = inactive_output_description(key)
-                if desc not in db_names:
-                    lines.append('output "{}" disable'.format(desc))
-                    cmds.append('output "{}" disable'.format(desc))
-
-        print("[Saving]")
-        for line in lines:
-            print(line)
-
-        # Check if the outputs file exists
-        if os.path.isfile(outputs_path):
-            # Load a backup to restore settings if needed
-            backup = load_text_file(outputs_path).splitlines()
-        else:
-            backup = []
-
-        save_list_to_text_file(lines, outputs_path)
-
-        print("[Executing]")
-        for cmd in cmds:
-            print(cmd)
-
-        i3 = Connection()
-        for cmd in cmds:
-            i3.command(cmd)
-
-        create_confirm_win(backup, outputs_path)
-
-    elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
-        transforms = {"normal": 0, "90": 1, "180": 2, "270": 3, "flipped": 4, "flipped-90": 5, "flipped-180": 6,
-                      "flipped-270": 7}
-        for db in display_buttons:
-            name = db.name if not use_desc else "desc:{}".format(db.description.replace("#", "##"))
-            db_names.append(name)
-
-            line = "monitor={},{}x{}@{},{}x{},{}".format(name, db.physical_width, db.physical_height, db.refresh, db.x, db.y, db.scale)
-            if db.mirror:
-                line += ",mirror,{}".format(db.mirror)
-            if db.ten_bit:
-                line += ",bitdepth,10"
-
-            lines.append(line)
-            if db.transform != "normal":
-                lines.append("monitor={},transform,{}".format(name, transforms[db.transform]))
-
-            # avoid looking up the hardware name
-            if db.name in outputs_activity and not outputs_activity[db.name]:
-                lines.append("monitor={},disable".format(name))
-
-            cmd = "on" if db.dpms else "off"
-            hyprctl(f"dispatch dpms {cmd} {db.name}")
-
-        print("[Saving]")
-        for line in lines:
-            print(line)
-
-        backup = []
-        if os.path.isfile(outputs_path):
-            backup = load_text_file(outputs_path).splitlines()
-        save_list_to_text_file(lines, outputs_path)
-        create_confirm_win(backup, outputs_path)
-
-def create_confirm_win(backup, path):
+def create_confirm_win(backup, path, config_dir=None, profile_name=None):
     global confirm_win
     if confirm_win:
         confirm_win.destroy()
@@ -939,7 +960,7 @@ def create_confirm_win(backup, path):
 
     grid.attach(btn_restore, 0, 2, 1, 1)
     btn_keep = Gtk.Button.new_with_label(voc["keep"])
-    btn_keep.connect("clicked", keep_current_settings)
+    btn_keep.connect("clicked", keep_current_settings, config_dir, profile_name)
     grid.attach(btn_keep, 1, 2, 1, 1)
 
     confirm_win.show_all()
@@ -958,14 +979,22 @@ def count_down(label, backup, path):
     restore_old_settings(None, backup, path)
 
 
-def keep_current_settings(btn):
+def keep_current_settings(btn, config_dir=None, profile_name=None):
     if src_tag > 0:
         GLib.Source.remove(src_tag)
     confirm_win.close()
 
+    if config_dir and profile_name:
+        if config.get("profile-bound-wallpapers", True):
+            threading.Thread(
+                target=WallpaperManager.apply_profile_wallpapers,
+                args=(config_dir, profile_name),
+                daemon=True,
+            ).start()
+
 
 def restore_old_settings(btn, backup, path):
-    print("Restoring old settings...")
+    print("[Settings] Restoring old settings...")
     if src_tag > 0:
         GLib.Source.remove(src_tag)
 
@@ -981,7 +1010,7 @@ def restore_old_settings(btn, backup, path):
         # remove "{"
         single_line = single_line.replace("{", "")
         # convert multiple spaces into single
-        single_line = ' '.join(single_line.split())
+        single_line = " ".join(single_line.split())
         cmds = single_line.split("}")
         # execute line by line
         i3 = Connection()
@@ -1001,50 +1030,65 @@ def restore_old_settings(btn, backup, path):
 
 
 def main():
-    GLib.set_prgname('nwg-displays')
+    GLib.set_prgname("nwg-displays")
 
     parser = argparse.ArgumentParser()
 
     if sway:
-        parser.add_argument("-o",
-                            "--outputs_path",
-                            type=str,
-                            default="{}/outputs".format(sway_config_dir),
-                            help="path to save Outputs config to, default: {}".format(
-                                "{}/outputs".format(sway_config_dir)))
+        parser.add_argument(
+            "-o",
+            "--outputs_path",
+            type=str,
+            default="{}/outputs".format(sway_config_dir),
+            help="path to save Outputs config to, default: {}".format(
+                "{}/outputs".format(sway_config_dir)
+            ),
+        )
 
-        parser.add_argument("-n",
-                            "--num_ws",
-                            type=int,
-                            default=8,
-                            help="number of Workspaces in use, default: 8")
+        parser.add_argument(
+            "-n",
+            "--num_ws",
+            type=int,
+            default=8,
+            help="number of Workspaces in use, default: 8",
+        )
 
     elif hypr:
-        parser.add_argument("-m",
-                            "--monitors_path",
-                            type=str,
-                            default="{}/monitors.conf".format(hypr_config_dir),
-                            help="path to save the monitors.conf file to, default: {}".format(
-                                "{}/monitors.conf".format(hypr_config_dir)))
+        parser.add_argument(
+            "-m",
+            "--monitors_path",
+            type=str,
+            default="{}/monitors.conf".format(hypr_config_dir),
+            help="path to save the monitors.conf file to, default: {}".format(
+                "{}/monitors.conf".format(hypr_config_dir)
+            ),
+        )
 
-        parser.add_argument("-w",
-                            "--workspaces_path",
-                            type=str,
-                            default="{}/workspaces.conf".format(hypr_config_dir),
-                            help="path to save the workspaces.conf file to, default: {}".format(
-                                "{}/workspaces.conf".format(hypr_config_dir)))
+        parser.add_argument(
+            "-w",
+            "--workspaces_path",
+            type=str,
+            default="{}/workspaces.conf".format(hypr_config_dir),
+            help="path to save the workspaces.conf file to, default: {}".format(
+                "{}/workspaces.conf".format(hypr_config_dir)
+            ),
+        )
 
-        parser.add_argument("-n",
-                            "--num_ws",
-                            type=int,
-                            default=10,
-                            help="number of Workspaces in use, default: 10")
+        parser.add_argument(
+            "-n",
+            "--num_ws",
+            type=int,
+            default=10,
+            help="number of Workspaces in use, default: 10",
+        )
 
-    parser.add_argument("-v",
-                        "--version",
-                        action="version",
-                        version="%(prog)s version {}".format(__version__),
-                        help="display version information")
+    parser.add_argument(
+        "-v",
+        "--version",
+        action="version",
+        version="%(prog)s version {}".format(__version__),
+        help="display version information",
+    )
     args = parser.parse_args()
 
     load_vocabulary()
@@ -1060,30 +1104,6 @@ def main():
     elif hypr:
         if os.path.isdir(hypr_config_dir):
             outputs_path = args.monitors_path
-            if os.path.lexists(outputs_path):
-                is_writable = os.access(outputs_path, os.W_OK)
-                if os.path.islink(outputs_path) and not is_writable:
-                    eprint(f"INFO: '{outputs_path}' is a read-only symlink. Replacing with a writable file.")
-                    tmp_path = f"{outputs_path}.tmp"
-                    try:
-                        with open(outputs_path, 'r') as src_file, open(tmp_path, 'w') as tmp_file:
-                            tmp_file.write(src_file.read())
-                        backup_path = f"{outputs_path}.bkp"
-                        counter = 1
-                        while os.path.lexists(backup_path):
-                            backup_path = f"{outputs_path}.bkp{counter}"
-                            counter += 1
-                        eprint(f"INFO: Backing up '{outputs_path}' to '{backup_path}'")
-                        os.rename(outputs_path, backup_path)
-                        os.rename(tmp_path, outputs_path)
-                    except Exception as e:
-                        eprint(f"ERROR: Failed to replace read-only symlink: {e}")
-                elif not os.path.islink(outputs_path) and not is_writable:
-                    eprint(f"INFO: '{outputs_path}' is a read-only file. Making it writable.")
-                    try:
-                        os.chmod(outputs_path, os.stat(outputs_path).st_mode | stat.S_IWUSR)
-                    except Exception as e:
-                        eprint(f"ERROR: Failed to make file writable: {e}")
             workspaces_path = args.workspaces_path
         else:
             eprint("Hyprland config directory not found!")
@@ -1093,19 +1113,19 @@ def main():
     global num_ws
     num_ws = args.num_ws
     if sway:
-        print("Number of workspaces: {}".format(num_ws))
+        print("[Info] Number of workspaces: {}".format(num_ws))
 
     config_file = os.path.join(config_dir, "config")
     global config
     if not os.path.isfile(config_file):
         # migrate old config file, if not yet migrated
         if os.path.isfile(os.path.join(old_config_dir, "config")):
-            print("Migrating config to the proper path...")
+            print("[Config] Migrating config to the proper path...")
             os.rename(old_config_dir, config_dir)
         else:
             if not os.path.isdir(config_dir):
                 os.makedirs(config_dir, exist_ok=True)
-            print("'{}' file not found, creating default".format(config_file))
+            print("[Config] '{}' file not found, creating default".format(config_file))
             save_json(config, config_file)
     else:
         config = load_json(config_file)
@@ -1113,7 +1133,13 @@ def main():
     if config_keys_missing(config, config_file):
         config = load_json(config_file)
 
-    eprint("Settings: {}".format(config))
+    if "profile-bound-wallpapers" not in config:
+        config["profile-bound-wallpapers"] = True
+        save_json(config, config_file)
+
+    # Initialize the profile manager
+    global profile_manager
+    profile_manager = ProfileManager(config_dir, config, voc)
 
     global snap_threshold_scaled
     snap_threshold_scaled = config["snap-threshold"]
@@ -1125,16 +1151,21 @@ def main():
     screen = Gdk.Screen.get_default()
     provider = Gtk.CssProvider()
     style_context = Gtk.StyleContext()
-    style_context.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
+    style_context.add_provider_for_screen(
+        screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
+    )
     try:
         file = os.path.join(dir_name, "resources/style.css")
         provider.load_from_path(file)
     except:
-        sys.stderr.write("ERROR: {} file not found, using GTK styling\n".format(os.path.join(dir_name,
-                                                                                             "resources/style.css")))
+        sys.stderr.write(
+            "ERROR: {} file not found, using GTK styling\n".format(
+                os.path.join(dir_name, "resources/style.css")
+            )
+        )
 
     window.connect("key-release-event", handle_keyboard)
-    window.connect('destroy', Gtk.main_quit)
+    window.connect("destroy", Gtk.main_quit)
 
     builder.get_object("lbl-modes").set_label("{}:".format(voc["modes"]))
     builder.get_object("lbl-position-x").set_label("{}:".format(voc["position-x"]))
@@ -1182,37 +1213,49 @@ def main():
     global form_view_scale
     form_view_scale = builder.get_object("view-scale")
     form_view_scale.set_tooltip_text(voc["view-scale-tooltip"])
-    adj = Gtk.Adjustment(lower=0.1, upper=0.6, step_increment=0.05, page_increment=0.1, page_size=0.1)
+    adj = Gtk.Adjustment(
+        lower=0.1, upper=0.6, step_increment=0.05, page_increment=0.1, page_size=0.1
+    )
     form_view_scale.configure(adj, 1, 2)
     form_view_scale.connect("changed", on_view_scale_changed)
 
     global form_x
     form_x = builder.get_object("x")
-    adj = Gtk.Adjustment(lower=0, upper=60000, step_increment=1, page_increment=10, page_size=1)
+    adj = Gtk.Adjustment(
+        lower=0, upper=60000, step_increment=1, page_increment=10, page_size=1
+    )
     form_x.configure(adj, 1, 0)
     form_x.connect("value-changed", on_pos_x_changed)
 
     global form_y
     form_y = builder.get_object("y")
-    adj = Gtk.Adjustment(lower=0, upper=40000, step_increment=1, page_increment=10, page_size=1)
+    adj = Gtk.Adjustment(
+        lower=0, upper=40000, step_increment=1, page_increment=10, page_size=1
+    )
     form_y.configure(adj, 1, 0)
     form_y.connect("value-changed", on_pos_y_changed)
 
     global form_width
     form_width = builder.get_object("width")
-    adj = Gtk.Adjustment(lower=0, upper=7680, step_increment=1, page_increment=10, page_size=1)
+    adj = Gtk.Adjustment(
+        lower=0, upper=7680, step_increment=1, page_increment=10, page_size=1
+    )
     form_width.configure(adj, 1, 0)
     form_width.connect("value-changed", on_width_changed)
 
     global form_height
     form_height = builder.get_object("height")
-    adj = Gtk.Adjustment(lower=0, upper=4320, step_increment=1, page_increment=10, page_size=1)
+    adj = Gtk.Adjustment(
+        lower=0, upper=4320, step_increment=1, page_increment=10, page_size=1
+    )
     form_height.configure(adj, 1, 0)
     form_height.connect("value-changed", on_height_changed)
 
     global form_scale
     form_scale = builder.get_object("scale")
-    adj = Gtk.Adjustment(lower=0.1, upper=10, step_increment=0.1, page_increment=10, page_size=1)
+    adj = Gtk.Adjustment(
+        lower=0.1, upper=10, step_increment=0.1, page_increment=10, page_size=1
+    )
     form_scale.configure(adj, 0.1, 6)
     form_scale.connect("value-changed", on_scale_changed)
 
@@ -1226,7 +1269,9 @@ def main():
 
     global form_refresh
     form_refresh = builder.get_object("refresh")
-    adj = Gtk.Adjustment(lower=1, upper=1200, step_increment=1, page_increment=10, page_size=1)
+    adj = Gtk.Adjustment(
+        lower=1, upper=1200, step_increment=1, page_increment=10, page_size=1
+    )
     form_refresh.configure(adj, 1, 3)
     form_refresh.connect("changed", on_refresh_changed)
 
@@ -1268,7 +1313,7 @@ def main():
     form_apply = builder.get_object("apply")
     form_apply.set_label(voc["apply"])
     if (sway and sway_config_dir) or (hypr and hypr_config_dir):
-        form_apply.connect("clicked", on_apply_button)
+        form_apply.connect("clicked", on_apply_button, profile_manager)
     else:
         form_apply.set_sensitive(False)
         form_apply.set_tooltip_text("Config dir not found")
@@ -1305,23 +1350,70 @@ def main():
     else:
         btn.destroy()
 
-    if hypr:
-        grid = builder.get_object("grid")
+    # Add profile management buttons
+    separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
+    form_wrapper_box.pack_start(separator, False, False, 10)
 
-        global form_ten_bit
-        form_ten_bit = Gtk.CheckButton.new_with_label(voc["10-bit-support"])
-        form_ten_bit.set_tooltip_text(voc["10-bit-support-tooltip"])
-        form_ten_bit.connect("toggled", on_ten_bit_toggled)
-        grid.attach(form_ten_bit, 5, 4, 1, 1)
+    profile_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
+    form_wrapper_box.pack_start(profile_box, False, False, 0)
 
-        lbl = Gtk.Label.new("Mirror:")
-        lbl.set_property("halign", Gtk.Align.END)
-        grid.attach(lbl, 6, 4, 1, 1)
+    # Add label for current profile name
+    profile_label = Gtk.Label()
+    profile_label.set_text(
+        voc.get("current-profile", "Profile") + ": " + voc.get("none", "None")
+    )
+    profile_label.set_property("margin-end", 10)
+    profile_label.set_property("margin-start", 5)
+    profile_box.pack_start(profile_label, False, False, 0)
 
-        global form_mirror
-        form_mirror = Gtk.ComboBoxText()
-        form_mirror.connect("changed", on_mirror_selected)
-        grid.attach(form_mirror, 7, 4, 1, 1)
+    btn_new_profile = Gtk.Button.new_with_label(voc.get("new", "New"))
+    btn_new_profile.set_tooltip_text(
+        voc.get("new-profile-tooltip", "Create a new profile")
+    )
+    btn_new_profile.connect("clicked", profile_manager.create_profile)
+    profile_box.pack_start(btn_new_profile, False, False, 0)
+
+    btn_select_profile = Gtk.Button.new_with_label(voc.get("select", "Select"))
+    btn_select_profile.set_tooltip_text(
+        voc.get("select-profile-tooltip", "Select a profile")
+    )
+    btn_select_profile.connect("clicked", profile_manager.select_profile)
+    profile_box.pack_start(btn_select_profile, False, False, 0)
+
+    btn_save_profile = Gtk.Button.new_with_label(voc.get("save", "Save"))
+    btn_save_profile.set_tooltip_text(
+        voc.get("save-profile-tooltip", "Save to current profile")
+    )
+    btn_save_profile.connect("clicked", profile_manager.save_profile)
+    btn_save_profile.set_sensitive(
+        False
+    )  # Initially disabled until a profile is selected
+    profile_box.pack_start(btn_save_profile, False, False, 0)
+
+    global form_profile_wallpapers
+    form_profile_wallpapers = Gtk.CheckButton.new_with_label(
+        voc.get("profile-bound-wallpapers", "Profile-bound wallpapers")
+    )
+    form_profile_wallpapers.set_tooltip_text(
+        voc.get(
+            "profile-bound-wallpapers-tooltip",
+            "Save and load wallpapers together with display profiles",
+        )
+    )
+    form_profile_wallpapers.set_active(config.get("profile-bound-wallpapers", True))
+    form_profile_wallpapers.connect("toggled", on_profile_wallpapers_toggled)
+    form_wrapper_box.pack_start(form_profile_wallpapers, False, False, 6)
+
+    # Register the save button with the profile manager
+    profile_manager.set_save_button(btn_save_profile)
+
+    # Register the profile label with the profile manager
+    profile_manager.set_profile_label(profile_label)
+
+    # Also pass display_buttons and other required data to profile manager
+    profile_manager.set_display_buttons(display_buttons)
+    profile_manager.set_fixed(fixed)
+    profile_manager.set_update_callback(update_form_from_widget)
 
     if display_buttons:
         update_form_from_widget(display_buttons[0])
@@ -1330,7 +1422,9 @@ def main():
     screen = Gdk.Screen.get_default()
     provider = Gtk.CssProvider()
     style_context = Gtk.StyleContext()
-    style_context.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
+    style_context.add_provider_for_screen(
+        screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
+    )
     css = b""" #popup { border-radius: 6px; border: solid 1px; border-color: #f00 } """
     provider.load_from_data(css)
 
@@ -1343,5 +1437,5 @@ def main():
     Gtk.main()
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     sys.exit(main())
